//==============================================================================
#include "socket_engine.h"
//------------------------------------------------------------------------------
#include <string.h>
//==============================================================================
bool prepare_server_node(se_basic_node * const obj, SOCKET s, SOCKADDR_IN sa)
{
    if (obj->_connected)
    {
        return false;
    }

    obj->s = s;
    obj->sa = sa;

    obj->_connected = true; 

    return true;
}
//------------------------------------------------------------------------------
bool prepare_client_node
(
    se_basic_node * const obj,
    const int af,
    const int type,
    const int protocol,
    const char *ip,
    const unsigned short port
)
{
    if (obj->_connected)
    {
        return false;
    }

    obj->af = af;
    obj->type = type;
    obj->protocol = protocol;

    obj->sa.sin_family = af;
    obj->sa.sin_port = htons(port);
    obj->sa.sin_addr.s_addr = inet_addr(ip);

    return true;
}
//==============================================================================
const char *se_basic_node::job_str[countof_jobs] =
{
    "no_job",
    "_connect",
    "_shutdown",
    "_disconnect",
    "_recv",
    "_send",
    "unknown_job"
};
//------------------------------------------------------------------------------
const char *se_basic_node::result_str[countof_results] =
{
    "no_result",
    "busy",
    "already_connected",
    "socket_failed",
    "can_not_enable_nonblocking_mode",
    "select_failed",
    "connection_attempt_failed",
    "already_disconnected",
    "shutdown_failed",
    "recv_failed",
    "connection_closed",
    "send_failed",
    "aborted_with_timeout",
    "aborted",
    "success",
    "success_with_overflow",
    "unknown_result"
};
//------------------------------------------------------------------------------
se_basic_node::se_basic_node(void):
    job_timeout(-1),
    package_mode(false),
    sizeof_package(0),
    sizeof_data_cur(0),
    sizeof_data(0),
    sizeof_buffer(0),
    buffer(0),
    af(AF_INET),
    type(SOCK_STREAM),
    protocol(IPPROTO_TCP),
    s(INVALID_SOCKET),
    _job_active(false),
    _job_activation_time(0),
    job(no_job),
    result(no_result),
    _ready_for_job_proc(false),
    _connected(false)
{
    memset(&sa, 0, sizeof(sa));
}
//------------------------------------------------------------------------------
se_basic_node::~se_basic_node(void)
{
    disconnect();
}
//------------------------------------------------------------------------------
bool se_basic_node::activate_job(const jobs job, const bool ready_for_job_proc)
{
    if (_job_active || job <= no_job || unknown_job <= job)
    {
        return false;
    }

    sizeof_package = 0;
    sizeof_data_cur = 0;
    
    _job_active = true;
    _job_activation_time = time(0);
    this->job = job;
    result = busy;
    _ready_for_job_proc = ready_for_job_proc;

    return true;
}
//------------------------------------------------------------------------------
void se_basic_node::abort_job(void)
{
    if (!_job_active)
    {
        return;
    }

    _job_active = false;
    result = aborted;
    result_proc();
}
//------------------------------------------------------------------------------
bool se_basic_node::job_active(void)
{
    return _job_active;
}
//------------------------------------------------------------------------------
std::time_t se_basic_node::job_activation_time(void)
{
    return _job_activation_time;
} 
//------------------------------------------------------------------------------
se_basic_node::jobs se_basic_node::get_job(void)
{
    return job;
}
//------------------------------------------------------------------------------
const char *se_basic_node::get_job_str(const jobs job)
{
    return job_str[job < no_job || unknown_job <= job? unknown_job : job];
}
//------------------------------------------------------------------------------
se_basic_node::results se_basic_node::get_result(void)
{
    return result;
}
//------------------------------------------------------------------------------
const char *se_basic_node::get_result_str(const results result)
{
    return result_str[result < no_result || unknown_result <= result? unknown_result : result];
}
//------------------------------------------------------------------------------
bool se_basic_node::ready_for_job_proc(void)
{
    return _ready_for_job_proc;
}
//------------------------------------------------------------------------------
void se_basic_node::job_proc(void)
{
    static char tmp[2048]; //       
    
    _ready_for_job_proc = false;
    
    if (!_job_active || job <= no_job || unknown_job <= job)
    {
        return;
    }

    if (job == _connect)
    {
        if (_connected)
        {
            _job_active = false;
            result = already_connected;
            result_proc();
        }
        else
        {
            if (s == INVALID_SOCKET)
            {
                s = socket(af, type, protocol);
                if (s == INVALID_SOCKET)
                {
                    _job_active = false;
                    result = socket_failed;
                    result_proc();
                }
                else
                {
                    if (!enable_nonblocking_mode(s, true))
                    {
                        closesocket(s);
                        s = INVALID_SOCKET;

                        _job_active = false;
                        result = can_not_enable_nonblocking_mode;
                        result_proc();
                    }
                    else
                    {
                        connect(s, (SOCKADDR *)&sa, sizeof(sa));
                    }
                }
            }
            else
            {
                fd_set w_set;
                FD_ZERO(&w_set);
                FD_SET(s, &w_set);
                fd_set e_set;
                FD_ZERO(&e_set);
                FD_SET(s, &e_set);
                timeval timeout = {0, 0};
                int select_res = select(s + 1, 0, &w_set, &e_set, &timeout);
                if (select_res == SOCKET_ERROR)
                {
                    shutdown(s, SD_BOTH);

                    enable_nonblocking_mode(s, false);
                    closesocket(s);
                    s = INVALID_SOCKET;

                    _job_active = false;
                    result = select_failed;
                    result_proc();
                }
                else if (select_res)
                {
                    if (FD_ISSET(s, &e_set))
                    {
                        shutdown(s, SD_BOTH);

                        enable_nonblocking_mode(s, false);
                        closesocket(s);
                        s = INVALID_SOCKET;

                        _job_active = false;
                        result = connection_attempt_failed;
                        result_proc();
                    }
                    else
                    {
                        _job_active = false;
                        result = success;
                        result_proc();

                        _connected = true;
                        after_connect();
                    }
                }
            }
        }
    }
    else if (job == _shutdown)
    {
        if (!_connected)
        {
            _job_active = false;
            result = already_disconnected;
            result_proc();
        }
        else
        {
            int shutdown_res = shutdown(s, SD_BOTH);

            _job_active = false;
            result = shutdown_res == SOCKET_ERROR? shutdown_failed : success;
            result_proc();
        }
    }
    else if (job == _disconnect)
    {
        if (!_connected)
        {
            _job_active = false;
            result = already_disconnected;
            result_proc();
        }
        else
        {
            enable_nonblocking_mode(s, false);
            closesocket(s);
            s = INVALID_SOCKET;

            _job_active = false;
            result = success;
            result_proc();

            _connected = false;
            after_disconnect();
        }
    }
    else if (job == _recv)
    {
        if (!_connected)
        {
            _job_active = false;
            result = already_disconnected;
            result_proc();
        }
        else
        {
            fd_set s_set;
            FD_ZERO(&s_set);
            FD_SET(s, &s_set);
            timeval timeout = {0, 0};
            int select_res = select(s + 1, &s_set, 0, 0, &timeout);
            if (select_res == SOCKET_ERROR)
            {
                _job_active = false;
                result = select_failed;
                result_proc();
            }
            else if (select_res)
            {
                char *buf;
                int len;

                if (sizeof_data_cur >= sizeof_buffer)
                {
                    buf = tmp;
                    len = sizeof(tmp);
                }
                else
                {
                    buf = buffer + sizeof_data_cur;
                    len = (sizeof_data > sizeof_buffer? sizeof_buffer : sizeof_data) - sizeof_data_cur;
                }

                sizeof_package = recv(s, buf, len, 0);
                if (sizeof_package == SOCKET_ERROR)
                {
                    _job_active = false;
                    result = recv_failed;
                    result_proc();
                }
                else if (!sizeof_package) //  
                {
                    enable_nonblocking_mode(s, false);
                    closesocket(s);
                    s = INVALID_SOCKET;

                    _job_active = false;
                    result = connection_closed;
                    result_proc();

                    _connected = false;
                    after_disconnect();
                }
                else
                {
                    after_package(buf, sizeof_package);

                    sizeof_data_cur += sizeof_package;
                    if (sizeof_data_cur == sizeof_data || package_mode)
                    {
                        _job_active = false;
                        result = sizeof_data > sizeof_buffer? success_with_overflow : success;
                        result_proc();
                    }
                }
            }
        }
    }
    else if (job == _send)
    {
        if (!_connected)
        {
            _job_active = false;
            result = already_disconnected;
            result_proc();
        }
        else
        {
            fd_set s_set;
            FD_ZERO(&s_set);
            FD_SET(s, &s_set);
            timeval timeout = {0, 0};
            int select_res = select(s + 1, 0, &s_set, 0, &timeout);
            if (select_res == SOCKET_ERROR)
            {
                _job_active = false;
                result = select_failed;
                result_proc();
            }
            else if (select_res)
            {
                char *buf;
                int len;

                buf = buffer + sizeof_data_cur;
                len = (sizeof_data > sizeof_buffer? sizeof_buffer : sizeof_data) - sizeof_data_cur;

                sizeof_package = send(s, buf, len, 0);
                if (sizeof_package == SOCKET_ERROR)
                {
                    _job_active = false;
                    result = send_failed;
                    result_proc();
                }
                else
                {
                    sizeof_data_cur += sizeof_package;
                    if (sizeof_data_cur == sizeof_data || sizeof_data_cur == sizeof_buffer)
                    {
                        _job_active = false;
                        result = sizeof_data > sizeof_buffer? success_with_overflow : success;
                        result_proc();
                    }
                }
            }
        }
    }

    if (_job_active && job_timeout != -1 && std::difftime(std::time(0), _job_activation_time) * 1000 >= job_timeout)
    {
        _job_active = false;
        result = aborted_with_timeout;
        result_proc();
    }
}
//------------------------------------------------------------------------------
std::string se_basic_node::get_ip(void)
{
    return inet_ntoa(sa.sin_addr);
}
//------------------------------------------------------------------------------
unsigned short se_basic_node::get_port(void)
{
    return ntohs(sa.sin_port);
}
//------------------------------------------------------------------------------
void se_basic_node::disconnect(void)
{
    abort_job();
    
    if (_connected)
    {
        shutdown(s, SD_BOTH);
        
        enable_nonblocking_mode(s, false);
        closesocket(s);
        s = INVALID_SOCKET;

        _connected = false;
        after_disconnect();
    }
}
//------------------------------------------------------------------------------
bool se_basic_node::connected(void)
{
    return _connected;
}
//==============================================================================
bool se_client_node::prepare
(
    const int af,
    const int type,
    const int protocol,
    const char *ip,
    const unsigned short port
)
{
    return
    prepare_client_node
    (
        this,
        af,
        type,
        protocol,
        ip,
        port
    );
}
//==============================================================================
const char *se_server::error_str[countof_errors] =
{
    "no_error",
    "socket_failed",
    "bind_failed",
    "listen_failed",
    "can_not_enable_nonblocking_mode",
    "accept_failed",
    "can_not_enable_nonblocking_mode_on_accepted_socket",
    "select_failed",
    "unknown_error"
};
//------------------------------------------------------------------------------
se_server::se_server(void):
    af(AF_INET),
    type(SOCK_STREAM),
    protocol(IPPROTO_TCP),
    s(INVALID_SOCKET),
    _active(false)
{
    memset(&sa, 0, sizeof(sa));
}
//------------------------------------------------------------------------------
se_server::~se_server(void)
{
    deactivate();
}
//------------------------------------------------------------------------------
bool se_server::prepare
(
    const int af,
    const int type,
    const int protocol,
    const char *ip,
    const unsigned short port
)
{
    if (_active)
    {
        return false;
    }

    this->af = af;
    this->type = type;
    this->protocol = protocol;

    sa.sin_family = af;
    sa.sin_port = htons(port);
    sa.sin_addr.s_addr = inet_addr(ip);

    return true;
}
//------------------------------------------------------------------------------
bool se_server::activate(const bool need_for_listen)
{
    if (_active)
    {
        return false;
    }

    s = socket(af, type, protocol);
    if (s == INVALID_SOCKET)
    {
        after_error(socket_failed);
        return false;
    }
    if (!enable_nonblocking_mode(s, true))
    {
        closesocket(s);
        s = INVALID_SOCKET;
        after_error(can_not_enable_nonblocking_mode);
        return false;
    }
    if (bind(s, (SOCKADDR *)&sa, sizeof(sa)) == SOCKET_ERROR)
    {
        closesocket(s);
        s = INVALID_SOCKET;
        after_error(bind_failed);
        return false;
    }
    if (need_for_listen && listen(s, SOMAXCONN) == SOCKET_ERROR)
    {
        closesocket(s);
        s = INVALID_SOCKET;
        after_error(listen_failed);
        return false;
    }

    _active = true;
    
    return true;
}
//------------------------------------------------------------------------------
void se_server::deactivate(void)
{
    if (!_active)
    {
        return;
    }

    for
    (
        std::list<se_server_node *>::iterator p = nodes.begin();
        p != nodes.end();
        p++
    )
    {
        (*p)->disconnect();
        delete *p;
    }
    nodes.clear();

    enable_nonblocking_mode(s, false);
    closesocket(s);
    s = INVALID_SOCKET;

    _active = false;
}
//------------------------------------------------------------------------------
bool se_server::active(void)
{
    return _active;
}
//------------------------------------------------------------------------------
void se_server::job_proc(void)
{
    if (!_active)
    {
        return;
    }

    fd_set s_set;
    FD_ZERO(&s_set);
    FD_SET(s, &s_set);
    timeval timeout = {0, 0};
    int select_res = select(s + 1, &s_set, 0, 0, &timeout);
    if (select_res == SOCKET_ERROR)
    {
        after_error(select_failed);
    }
    else if (select_res == 1)
    {
        SOCKET ns;

        SOCKADDR_IN nsa;
        socklen_t sizeof_nsa = sizeof(nsa);

        ns = accept(s, (SOCKADDR *)&nsa, &sizeof_nsa);
        if (ns == INVALID_SOCKET)
        {
            after_error(accept_failed);
        }
        else
        {
            if (!enable_nonblocking_mode(ns, true))
            {
                closesocket(ns);
                after_error(can_not_enable_nonblocking_mode_on_accepted_socket);
            }
            else
            {
                se_server_node *node = new_node();
                prepare_server_node(node, ns, nsa);
                nodes.push_back(node);
                node->after_connect();
            }
        }
    }

    for
    (
        std::list<se_server_node *>::iterator p = nodes.begin();
        p != nodes.end();
    )
    {
        while (true)
        {
            (*p)->before_job_proc();
            (*p)->job_proc();
            if (!(*p)->ready_for_job_proc()) break;
        }
        if (!(*p)->connected())
        {
            delete *p;
            p = nodes.erase(p);
        }
        else
        {
            p++;
        }
    }
}
//------------------------------------------------------------------------------
const char *se_server::get_error_str(const errors error)
{
    return error_str[error < no_error || unknown_error <= error? unknown_error : error];
}
//------------------------------------------------------------------------------
std::string se_server::get_ip(void)
{
    return inet_ntoa(sa.sin_addr);
}
//------------------------------------------------------------------------------
unsigned short se_server::get_port(void)
{
    return ntohs(sa.sin_port);
}
//------------------------------------------------------------------------------
se_server_node *se_server::new_node(void)
{
    return new se_server_node();
}
//==============================================================================
se_client::se_client(void)
{}
//------------------------------------------------------------------------------
se_client::~se_client(void)
{
    disconnect();
    delete_all_nodes();
}
//------------------------------------------------------------------------------
void se_client::job_proc(void)
{
    for
    (
        std::list<se_client_node *>::iterator p = nodes.begin();
        p != nodes.end();
        p++
    )
    {
        while (true)
        {
            (*p)->before_job_proc();
            (*p)->job_proc();
            if (!(*p)->ready_for_job_proc()) break;
        }
    }
}
//------------------------------------------------------------------------------
void se_client::disconnect(void)
{
    for
    (
        std::list<se_client_node *>::iterator p = nodes.begin();
        p != nodes.end();
        p++
    )
    {
        (*p)->disconnect();
    }
}
//------------------------------------------------------------------------------
void se_client::delete_all_nodes(void)
{
    for
    (
        std::list<se_client_node *>::iterator p = nodes.begin();
        p != nodes.end();
        p++
    )
    {
        delete *p;
    }
    nodes.clear();
}

